import ast
import torch
import torch.nn as nn
import torch.nn.functional as F
from dhgbench.lib_models.HNN.mlp import MLP
from dhgbench.lib_models.HNN.ehnn_linear import EHNNLinearConv
from dhgbench.lib_models.HNN.ehnn_transformer import EHNNTransformerConv

class EHNN(nn.Module):
    
    def __init__(self, num_features, num_targets, args, ehnn_cache):
        super().__init__()
        edge_orders = ehnn_cache["edge_orders"]  # [|E|,]
        overlaps = ehnn_cache["overlaps"]  # [|overlaps|,]
        max_edge_order = int(edge_orders.max().item())
        max_overlap = int(overlaps.max().item()) if overlaps is not None else 0
        hypernet_info = (max_edge_order, max_edge_order, max_overlap)
        
        self.num_layers = args.All_num_layers
        self.hidden_dim = args.ehnn_hidden_channel
        self.ehnn_type=args.ehnn_type

        self.dropout=args.dropout
        self.decoder_hidden=args.decoder_hidden
        self.decoder_num_layers =args.decoder_num_layers
        self.decoder_norm=args.normalization 

        self.convs = nn.ModuleList()
        if self.ehnn_type == "linear":
            if self.num_layers == 1:
                self.convs.append(EHNNLinearConv(num_features,self.hidden_dim,
                    args.ehnn_inner_channel,args.dropout,hypernet_info,args.ehnn_pe_dim,
                    args.ehnn_hyper_dim,args.ehnn_hyper_layers,
                    args.ehnn_hyper_dropout,ast.literal_eval(args.ehnn_force_broadcast),
                    args.ehnn_input_dropout,ast.literal_eval(args.ehnn_mlp_classifier)))
            else:
                self.convs.append(EHNNLinearConv(num_features,self.hidden_dim,
                    args.ehnn_inner_channel,args.dropout,hypernet_info,args.ehnn_pe_dim,
                    args.ehnn_hyper_dim,args.ehnn_hyper_layers,
                    args.ehnn_hyper_dropout,ast.literal_eval(args.ehnn_force_broadcast),
                    args.ehnn_input_dropout,ast.literal_eval(args.ehnn_mlp_classifier)))
                for _ in range(self.num_layers-2):
                    self.convs.append(EHNNLinearConv(self.hidden_dim,self.hidden_dim,
                    args.ehnn_inner_channel,args.dropout,hypernet_info,args.ehnn_pe_dim,
                    args.ehnn_hyper_dim,args.ehnn_hyper_layers,
                    args.ehnn_hyper_dropout,ast.literal_eval(args.ehnn_force_broadcast),
                    args.ehnn_input_dropout,ast.literal_eval(args.ehnn_mlp_classifier)))
                # Output heads is set to 1 as default
                self.convs.append(EHNNLinearConv(self.hidden_dim,self.hidden_dim,
                    args.ehnn_inner_channel,args.dropout,hypernet_info,args.ehnn_pe_dim,
                    args.ehnn_hyper_dim,args.ehnn_hyper_layers,
                    args.ehnn_hyper_dropout,ast.literal_eval(args.ehnn_force_broadcast),
                    args.ehnn_input_dropout,ast.literal_eval(args.ehnn_mlp_classifier)))
            
        elif self.ehnn_type == "transformer":
            if self.num_layers == 1:
                self.convs.append(EHNNTransformerConv(num_features,self.hidden_dim,
                    args.ehnn_qk_channel,args.ehnn_hidden_channel,args.ehnn_n_heads,
                    args.dropout,hypernet_info,args.ehnn_inner_channel,args.ehnn_pe_dim,args.ehnn_hyper_dim,
                    args.ehnn_hyper_layers,args.ehnn_hyper_dropout,ast.literal_eval(args.ehnn_force_broadcast),
                    args.ehnn_input_dropout,ast.literal_eval(args.ehnn_mlp_classifier),
                    args.ehnn_att0_dropout,args.ehnn_att1_dropout))
            else:
                self.convs.append(EHNNTransformerConv(num_features,self.hidden_dim,
                    args.ehnn_qk_channel,args.ehnn_hidden_channel,args.ehnn_n_heads,
                    args.dropout,hypernet_info,args.ehnn_inner_channel,args.ehnn_pe_dim,args.ehnn_hyper_dim,
                    args.ehnn_hyper_layers,args.ehnn_hyper_dropout,ast.literal_eval(args.ehnn_force_broadcast),
                    args.ehnn_input_dropout,ast.literal_eval(args.ehnn_mlp_classifier),
                    args.ehnn_att0_dropout,args.ehnn_att1_dropout))
                for _ in range(self.num_layers-2):
                    self.convs.append(EHNNTransformerConv(self.hidden_dim,self.hidden_dim,
                    args.ehnn_qk_channel,args.ehnn_hidden_channel,args.ehnn_n_heads,
                    args.dropout,hypernet_info,args.ehnn_inner_channel,args.ehnn_pe_dim,args.ehnn_hyper_dim,
                    args.ehnn_hyper_layers,args.ehnn_hyper_dropout,ast.literal_eval(args.ehnn_force_broadcast),
                    args.ehnn_input_dropout,ast.literal_eval(args.ehnn_mlp_classifier),
                    args.ehnn_att0_dropout,args.ehnn_att1_dropout))
                self.convs.append(EHNNTransformerConv(self.hidden_dim,self.hidden_dim,
                    args.ehnn_qk_channel,args.ehnn_hidden_channel,args.ehnn_n_heads,
                    args.dropout,hypernet_info,args.ehnn_inner_channel,args.ehnn_pe_dim,args.ehnn_hyper_dim,
                    args.ehnn_hyper_layers,args.ehnn_hyper_dropout,ast.literal_eval(args.ehnn_force_broadcast),
                    args.ehnn_input_dropout,ast.literal_eval(args.ehnn_mlp_classifier),
                    args.ehnn_att0_dropout,args.ehnn_att1_dropout))
        else:
            raise NotImplementedError

        self.decoder = MLP(
            in_channels=self.hidden_dim,
            hidden_channels=self.decoder_hidden,
            out_channels=num_targets,
            num_layers=self.decoder_num_layers,
            dropout=self.dropout,
            Normalization=self.decoder_norm,
            InputNorm=False,
        )

    def reset_parameters(self):
        for conv in self.convs:
            conv.reset_parameters()
        self.decoder.reset_parameters()

    def forward(self, data):
        """forward method
        :param data:
        :param ehnn_cache:
        :return: [N, C] dense
        """
        x = data.x
        ehnn_cache = data.ehnn_cache
        
        if self.ehnn_type == 'transformer':
            with torch.no_grad():
                ehnn_cache["incidence"] = ehnn_cache["incidence"].coalesce()
                n, e = ehnn_cache["incidence"].size()
                node_indices = torch.arange(0, n, device=ehnn_cache["incidence"].device)[
                    None, :
                ].repeat(2, 1)
                node_indices[1] += e
                ehnn_cache["indices_with_nodes"] = torch.cat(
                    (ehnn_cache["incidence"].indices(), node_indices), dim=1
                )
        
        for i, conv in enumerate(self.convs[:-1]):
            x,_ = conv(x,ehnn_cache) 
            x = F.dropout(x, p=self.dropout, training=self.training)

        x,x_e = self.convs[-1](x, ehnn_cache)
        x = self.decoder(x)

        return x,x_e




import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

class MLP(nn.Module):
    """ adapted from https://github.com/CUAI/CorrectAndSmooth/blob/master/gen_models.py """

    def __init__(self, in_channels, hidden_channels, out_channels, num_layers,
                 dropout=.5, Normalization='bn', InputNorm=False):
        super(MLP, self).__init__()

        self.in_channels = in_channels
        self.hidden_channels = hidden_channels
        self.out_channels = out_channels

        self.lins = nn.ModuleList()
        self.normalizations = nn.ModuleList()
        self.InputNorm = InputNorm

        assert Normalization in ['bn', 'ln', 'None']
        if Normalization == 'bn':
            if num_layers == 1:
                # just linear layer i.e. logistic regression
                if InputNorm:
                    self.normalizations.append(nn.BatchNorm1d(in_channels))
                else:
                    self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, out_channels))
            else:
                if InputNorm:
                    self.normalizations.append(nn.BatchNorm1d(in_channels))
                else:
                    self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, hidden_channels))
                self.normalizations.append(nn.BatchNorm1d(hidden_channels))
                for _ in range(num_layers - 2):
                    self.lins.append(
                        nn.Linear(hidden_channels, hidden_channels))
                    self.normalizations.append(nn.BatchNorm1d(hidden_channels))
                self.lins.append(nn.Linear(hidden_channels, out_channels))
        elif Normalization == 'ln':
            if num_layers == 1:
                # just linear layer i.e. logistic regression
                if InputNorm:
                    self.normalizations.append(nn.LayerNorm(in_channels))
                else:
                    self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, out_channels))
            else:
                if InputNorm:
                    self.normalizations.append(nn.LayerNorm(in_channels))
                else:
                    self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, hidden_channels))
                self.normalizations.append(nn.LayerNorm(hidden_channels))
                for _ in range(num_layers - 2):
                    self.lins.append(
                        nn.Linear(hidden_channels, hidden_channels))
                    self.normalizations.append(nn.LayerNorm(hidden_channels))
                self.lins.append(nn.Linear(hidden_channels, out_channels))
        else:
            if num_layers == 1:
                # just linear layer i.e. logistic regression
                self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, out_channels))
            else:
                self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, hidden_channels))
                self.normalizations.append(nn.Identity())
                for _ in range(num_layers - 2):
                    self.lins.append(
                        nn.Linear(hidden_channels, hidden_channels))
                    self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(hidden_channels, out_channels))

        self.dropout = dropout

    def reset_parameters(self):
        for lin in self.lins:
            lin.reset_parameters()
        for normalization in self.normalizations:
            if not (normalization.__class__.__name__ == 'Identity'):
                normalization.reset_parameters()

    def forward(self, x):
        x = self.normalizations[0](x)
        for i, lin in enumerate(self.lins[:-1]):
            x = lin(x)
            x = F.relu(x, inplace=True)
            x = self.normalizations[i+1](x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.lins[-1](x)
        return x

    def flops(self, x):
        num_samples = np.prod(x.shape[:-1])
        flops = num_samples * self.in_channels # first normalization
        flops += num_samples * self.in_channels * self.hidden_channels # first linear layer
        flops += num_samples * self.hidden_channels # first relu layer

        # flops for each layer
        per_layer = num_samples * self.hidden_channels * self.hidden_channels
        per_layer += num_samples * self.hidden_channels # relu + normalization
        flops += per_layer * (len(self.lins) - 2)

        flops += num_samples * self.out_channels * self.hidden_channels # last linear layer

        return flops

class PlainMLP(nn.Module):
    
    """ adapted from https://github.com/CUAI/CorrectAndSmooth/blob/master/gen_models.py """

    def __init__(self, in_channels, out_channels, args):
        super(PlainMLP, self).__init__()

        in_channels = in_channels
        hidden_channels = args.MLP_hidden
        out_channels = out_channels
        
        num_layers = args.All_num_layers
        Normalization = args.normalization

        self.lins = nn.ModuleList()
        self.normalizations = nn.ModuleList()
        InputNorm = args.InputNorm
        self.dropout = args.dropout

        assert Normalization in ['bn', 'ln', 'None']
        if Normalization == 'bn':
            if num_layers == 1:
                # just linear layer i.e. logistic regression
                if InputNorm:
                    self.normalizations.append(nn.BatchNorm1d(in_channels))
                else:
                    self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, out_channels))
            else:
                if InputNorm:
                    self.normalizations.append(nn.BatchNorm1d(in_channels))
                else:
                    self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, hidden_channels))
                self.normalizations.append(nn.BatchNorm1d(hidden_channels))
                for _ in range(num_layers - 2):
                    self.lins.append(
                        nn.Linear(hidden_channels, hidden_channels))
                    self.normalizations.append(nn.BatchNorm1d(hidden_channels))
                self.lins.append(nn.Linear(hidden_channels, out_channels))
        elif Normalization == 'ln':
            if num_layers == 1:
                # just linear layer i.e. logistic regression
                if InputNorm:
                    self.normalizations.append(nn.LayerNorm(in_channels))
                else:
                    self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, out_channels))
            else:
                if InputNorm:
                    self.normalizations.append(nn.LayerNorm(in_channels))
                else:
                    self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, hidden_channels))
                self.normalizations.append(nn.LayerNorm(hidden_channels))
                for _ in range(num_layers - 2):
                    self.lins.append(
                        nn.Linear(hidden_channels, hidden_channels))
                    self.normalizations.append(nn.LayerNorm(hidden_channels))
                self.lins.append(nn.Linear(hidden_channels, out_channels))
        else:
            if num_layers == 1:
                # just linear layer i.e. logistic regression
                self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, out_channels))
            else:
                self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(in_channels, hidden_channels))
                self.normalizations.append(nn.Identity())
                for _ in range(num_layers - 2):
                    self.lins.append(
                        nn.Linear(hidden_channels, hidden_channels))
                    self.normalizations.append(nn.Identity())
                self.lins.append(nn.Linear(hidden_channels, out_channels))

    def reset_parameters(self):
        for lin in self.lins:
            lin.reset_parameters()
        for normalization in self.normalizations:
            if not (normalization.__class__.__name__ == 'Identity'):
                normalization.reset_parameters()

    def forward(self, data):
        x = data.x 
        x = self.normalizations[0](x)
        for i, lin in enumerate(self.lins[:-1]):
            x = lin(x)
            x = F.relu(x, inplace=True)
            x = self.normalizations[i+1](x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.lins[-1](x)
        return x, None 

from typing import List

class MLP_inner(nn.Module):
    def __init__(self, dim_in, dim_out, dim_hidden, n_layers, dropout):
        super().__init__()
        self.n_layers = n_layers
        self.f = nn.ReLU(inplace=True)
        self.dropout = nn.Dropout(dropout)

        assert n_layers > 0
        self.layers = nn.ModuleList()
        if n_layers == 1:
            self.layers.append(nn.Linear(dim_in, dim_out))
        else:
            for layer_idx in range(n_layers):
                self.layers.append(
                    nn.Linear(
                        dim_hidden if layer_idx > 0 else dim_in,
                        dim_hidden if layer_idx < n_layers - 1 else dim_out,
                    )
                )

        self.norms = nn.ModuleList()
        for layer_idx in range(n_layers - 1):
            self.norms.append(nn.LayerNorm(dim_hidden if layer_idx > 0 else dim_in))

    def reset_parameters(self):
        for layer in self.layers:
            layer.reset_parameters()
        for norm in self.norms:
            norm.reset_parameters()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for idx in range(self.n_layers - 1):
            # x = self.norms[idx](x)
            x = self.layers[idx](x)
            x = self.f(x)
            x = self.dropout(x)
        x = self.layers[-1](x)
        return x

    def multi_forward(self, xs: List) -> List:
        for x in xs:
            assert len(x.size()) == 2
        return self.forward(torch.cat(xs)).split([x.size(0) for x in xs])


import torch
from torch_geometric.data import DataLoader

def generate_split_hypergraphs(dataset,train_ratio,val_ratio,seed):
    train_size = int(train_ratio * len(dataset))
    val_size = int(val_ratio * len(dataset))
    test_size = len(dataset) - train_size - val_size 
    train_set, val_set, test_set = torch.utils.data.random_split(dataset, [train_size, val_size, test_size], 
                                                                 generator=torch.Generator().manual_seed(42+seed))
    return train_set,val_set,test_set

def generate_hg_loaders(train_set,val_set,test_set,args):
    batch_size = args.hg_batch_size
    train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True)
    val_loader = DataLoader(val_set, batch_size=batch_size, shuffle=False)
    test_loader = DataLoader(test_set, batch_size=batch_size, shuffle=False)
    
    batch_loaders = {'train':train_loader,
                     'val':val_loader,
                     'test':test_loader}
    
    return batch_loaders


from .hgnn import HCHA
from .hnhn import HNHN
from .hypergcn import HyperGCN
from .allset import SetGNN
from .unignn import UniGNN
from .unigcn2 import UniGCNII
from .legcn import LEGCN
from .hypernd import HyperND
from .edgnn import EquivSetGNN
from .unigencoder import PlainUnigencoder
from .hjrl import HJRL
from .sheafhypergnn import SheafHyperGNN
from .ehnn import EHNN
from .tmphn import TMPHN
from .phenomnns import PhenomNNS
from .phenomnn import PhenomNN
from .dphgnn import DPHGNN
from .tfhnn import TFHNN
from .mlp import MLP,PlainMLP
from .hypergt import HyperGT
from .cegnn import CEGCN,CEGAT


from dhgbench.lib_utils.exp_agent import ExpAgent
from dhgbench.lib_models.HNN.preprocessing import algo_preprocessing
from dhgbench.lib_dataset.data_base import HyperDataset
from dhgbench.lib_dataset.preprocessing import data_processing
from dhgbench.parameter_parser import parameter_parser,method_config,set_task_args
from dhgbench.lib_dataset.data_perturbation import perturbation

def main():
    args = parameter_parser() 
    args = method_config(args) 
    args = set_task_args(args) 

    data=HyperDataset(args) 

    if args.task_type == 'hg_cls':
        data = data.multi_hypergraphs 
    else:
        data = data_processing(args,data)
        data._initialization_()

        if args.is_perturbed:
            if isinstance(args.pert_p,str):
                args.pert_p = eval(args.pert_p)
            if args.pert_mode not in ['spar_label','flip_label']:
                print('Robustness Perturbation for Structure and Feature')
                if args.is_poison:
                    data = perturbation(data,mode=args.pert_mode,p=args.pert_p,masks=None)
                else:
                    evasion_data = perturbation(data,mode=args.pert_mode,p=args.pert_p,masks=None)
            else:
                print('Robustness Perturbation for Supervision Signal')
                if args.is_poison == False:
                    raise ValueError('Label attack is expected to be the poison attack!')

        data = algo_preprocessing(data,args)

        if args.is_perturbed and not args.is_poison:
            evasion_data = algo_preprocessing(evasion_data,args)
            data.evasion_data = evasion_data 
    
    agent = ExpAgent(args)
    agent.running(args.task_type,data)

if __name__ == "__main__":
    main()


import torch
import pickle
import numpy as np
import os
import os.path as osp
from scipy.sparse import csr_matrix
from torch.utils.data import random_split
from dhgbench.lib_dataset.preprocessing import *
from dhgbench.lib_dataset.convert_datasets_to_pygDataset import dataset_Hypergraph
from torch_geometric.data import Data

import warnings
warnings.filterwarnings('ignore')

class HyperDataset(object):
    
    def __init__(self, args):

        self.device = args.device 
        self.args=args
        self.name=args.dname 
        self.method=args.method 
        
        self.trad_list=['cora', 'citeseer', 'pubmed',
                        'coauthor_cora', 'coauthor_dblp',
                        '20newsW100', 'ModelNet40','NTU2012', 
                        'Mushroom','zoo',
                        'yelp','walmart-trips-100','house-committees-100','walmart-trips','house-committees']
        
        self.large_list=['amazon_review','magpm','magpm_mini','trivago','ogbn_mag']
        self.fair_list = ['bail','credit','german']
        self.hete_list = ['actor','amazon','pokec','twitch']
        self.hg_list = ["RHG_3", "RHG_10", "RHG_table", "RHG_pyramid",
                        "IMDB_dir_form", "IMDB_dir_genre",
                        "IMDB_wri_form", "IMDB_wri_genre",
                        "stream_player","twitter_friend"] 
        
        # self.trad_path='../data/trad_data'
        # self.large_path='../data/large_data/' 
        # self.fair_path='../data/fair_data/'  
        # self.hete_path = '../data/hete_data/'
        # self.hg_path = '../data/hgcls_data/'
        data_root = args.data_dir.rstrip("/")

        self.trad_path = osp.join(data_root, "trad_data")
        self.large_path = osp.join(data_root, "large_data")
        self.fair_path = osp.join(data_root, "fair_data")
        self.hete_path = osp.join(data_root, "hete_data")
        self.hg_path = osp.join(data_root, "hgcls_data")

        self.pyg_root = osp.join(data_root, "pyg_data", "hypergraph_dataset_updated")

        if any(self.name in container for container in (self.trad_list,self.fair_list,self.hete_list,self.large_list)):
            self.load_data()
        elif self.name in self.hg_list:
            self.load_multi_data()
        else:
            raise ValueError('Undefined data')

    # -------------------------- Hypergraph-level Task Dataset Loading ------------------------
    
    def load_multi_data(self):
    
        if self.name in ["RHG_3", "RHG_10", "RHG_table", "RHG_pyramid"]:
            folder = "RHG"
        elif self.name in ["stream_player"]:
            folder = "STEAM"
        elif self.name in ["IMDB_dir_form", "IMDB_dir_genre"]:
            folder = "IMDB"
        elif self.name in ["IMDB_wri_form", "IMDB_wri_genre"]:
            folder = "IMDB"
        elif self.name in ["twitter_friend"]:
            folder = "TWITTER"
        else:
            raise NotImplementedError

        # read data
        x_list = []
        with open(f"{self.hg_path}/{folder}/{self.name}.txt", "r") as f:
            n_g = int(f.readline().strip())
            for _ in range(n_g):
                
                row = f.readline().strip().split()
                num_v, num_e = int(row[0]), int(row[1])
                g_lbl = [int(x) for x in row[2:]]
                v_lbl = f.readline().strip().split()
                v_lbl = [[int(x) for x in s.split("/")] for s in v_lbl]
                e_list = []
                for _ in range(num_e):#
                    row = f.readline().strip().split()
                    e_list.append([int(x) for x in row])
                
                hyperedge_index = convert_to_hyperedge_index(e_list) 
                v_deg=compute_degree_list(hyperedge_index)
                
                x_list.append(
                        {
                            "num_v": num_v,
                            "num_e": num_e,
                            "v_lbl": v_deg,
                            "g_lbl": g_lbl,
                            "edge_index": hyperedge_index,
                        }
                    )        

            v_lbl_set, g_lbl_set = set(), set()
            for x in x_list:
                if isinstance(x["v_lbl"][0], list):
                    for v_lbl in x["v_lbl"]:
                        v_lbl_set.update(v_lbl)
                else:
                    v_lbl_set.update(x["v_lbl"])
                g_lbl_set.update(x["g_lbl"])
                
            v_lbl_map = {x: i for i, x in enumerate(sorted(v_lbl_set))}
            g_lbl_map = {x: i for i, x in enumerate(sorted(g_lbl_set))}
            ft_dim, n_classes = len(v_lbl_set), len(g_lbl_set)
            for x in x_list:
                x["g_lbl"] = [g_lbl_map[c] for c in x["g_lbl"]]
                if isinstance(x["v_lbl"][0], list):
                    x["v_lbl"] = [tuple(sorted([v_lbl_map[c] for c in s])) for s in x["v_lbl"]]
                else:
                    x["v_lbl"] = [v_lbl_map[c] for c in x["v_lbl"]]
                x["v_ft"] = np.zeros((x["num_v"], ft_dim))
                row_idx, col_idx = [], []
                for v_idx, v_lbls in enumerate(x["v_lbl"]):
                    if isinstance(v_lbls, list) or isinstance(v_lbls, tuple):
                        for v_lbl in v_lbls:
                            row_idx.append(v_idx)
                            col_idx.append(v_lbl)
                    else:
                        row_idx.append(v_idx)
                        col_idx.append(v_lbls)
                x["v_ft"][row_idx, col_idx] = 1
            y_list = []
            y_list = [g["g_lbl"][0] for g in x_list]
            
        hgs=[] 
        for i,hg in enumerate(x_list):
            hypergraph = Data(x = torch.FloatTensor(hg["v_ft"]), hyperedge_index = hg["edge_index"], y = torch.LongTensor([y_list[i]]),
                            num_nodes = hg['num_v'],num_edges = hg['num_e'])
            hgs.append(hypergraph)
            
        self.multi_hypergraphs = MultiHyperaphDatasets(x=hgs, y=torch.LongTensor(y_list), 
                                                    num_hypergraphs=n_g, num_features=ft_dim, num_classes=n_classes)
    
    # -------------------------- Hypernode/edge-level Task Dataset Loading ------------------------
    
    def load_data(self):
        
        dname=self.name
        self.sens=None 
        self.norm=None 
        
        if dname in self.trad_list:
            
            path=self.trad_path
            if dname in ['cora', 'citeseer','pubmed']:
                # p2raw = osp.join(path, 'cocitation','')
                p2raw = osp.join(path, 'cocitation')
            elif dname in ['coauthor_cora', 'coauthor_dblp']:
                # p2raw = osp.join(path, 'coauthorship','')
                p2raw = osp.join(path, 'coauthorship')
            elif dname in ['yelp']:
                # p2raw = osp.join(path, 'yelp','')
                p2raw = osp.join(path, 'yelp')
            else:
                p2raw = path
            
            f_noise = self.args.feature_noise
            if (f_noise is not None) and dname in ['walmart-trips-100','house-committees-100','walmart-trips', 'house-committees']:
                # dataset = dataset_Hypergraph(name=dname,root = '../data/pyg_data/hypergraph_dataset_updated/',p2raw = p2raw,
                #                             feature_noise=f_noise)
                dataset = dataset_Hypergraph(name=dname, root=self.pyg_root, p2raw=p2raw)

            else:
                # dataset = dataset_Hypergraph(name=dname,root = '../data/pyg_data/hypergraph_dataset_updated/',p2raw = p2raw)
                dataset = dataset_Hypergraph(name=dname, root=self.pyg_root, p2raw=p2raw)

                
        elif dname in self.fair_list:
            
            p2raw = self.fair_path
            dataset = dataset_Hypergraph(name=dname,root = '../data/pyg_data/hypergraph_dataset_updated/',
                                            p2raw = p2raw)
            if dname=="credit":
                sens_idx = 1
            else:
                sens_idx = 0
            self.sens = dataset.data.x[:,sens_idx] 
        
        elif dname in self.hete_list:

            p2raw = self.hete_path
            dataset = dataset_Hypergraph(name=dname,root = '../data/pyg_data/hypergraph_dataset_updated/',
                                            p2raw = p2raw)

        elif dname in self.large_list:
        
            p2raw = self.large_path
            dataset = dataset_Hypergraph(name=dname,root = '../data/pyg_data/hypergraph_dataset_updated/',
                                            p2raw = p2raw)
        
        else:
            raise ValueError('Undefined data')
        
        data = dataset.data
        
        if dname in ['yelp', 'walmart-trips', 'house-committees', 'walmart-trips-100', 'house-committees-100','zoo']:
            #   Shift the y label to start with 0
            data.y = data.y - data.y.min()
        if not hasattr(data, 'n_x'):
            data.n_x = torch.tensor([data.x.shape[0]])
        if not hasattr(data, 'num_hyperedges'):
            # note that we assume the he_id is consecutive.
            data.num_hyperedges = torch.tensor(
                [data.edge_index[0].max()-data.n_x+1])

        self.data=data 
        
        self.x=data.x
        self.hyperedge_index=data.edge_index
        self.y=data.y

    def _initialization_(self):
        
        self.num_classes= len(self.y.unique())
        self.num_features=self.x.shape[1]
        self.num_nodes=self.x.shape[0]
        self.num_hyperedges=len(self.hyperedge_index[1].unique())
 
    def calculate_memory_size(self):

        labels_size = self.y.numel() * self.y.element_size()  
        features_size = self.x.numel() * self.x.element_size()
        hyperedge_index_size = self.hyperedge_index.numel() * self.hyperedge_index.element_size()

        total_size_bytes = labels_size + features_size + hyperedge_index_size
        total_size_gb = total_size_bytes / (1024 ** 3)  

        print(f"The memory cost of hypergraph: {total_size_bytes} Byte")
        print(f"The memory cost of hypergraph: {total_size_gb:.2f} GB")

    def to(self, device: str):
        
        self.x = self.x.to(device)
        self.hyperedge_index = self.hyperedge_index.to(device)
        self.y = self.y.to(device)
            
        return self

    def generate_random_split(self, train_ratio: float = 0.1, val_ratio: float = 0.1,
                              seed = None):

        num_train = int(self.num_nodes * train_ratio)
        num_val = int(self.num_nodes * val_ratio)
        num_test = self.num_nodes - (num_train + num_val)

        if seed is not None:
            generator = torch.Generator().manual_seed(seed)
        else:
            generator = torch.default_generator

        train_set, val_set, test_set = random_split(
            torch.arange(0, self.num_nodes), (num_train, num_val, num_test), 
            generator=generator)
        train_idx, val_idx, test_idx = \
            train_set.indices, val_set.indices, test_set.indices
        train_mask = torch.zeros((self.num_nodes,), device=self.device).to(torch.bool)
        val_mask = torch.zeros((self.num_nodes,), device=self.device).to(torch.bool)
        test_mask = torch.zeros((self.num_nodes,), device=self.device).to(torch.bool)

        train_mask[train_idx] = True
        val_mask[val_idx] = True
        test_mask[test_idx] = True

        '''
        [tensor([False, False, False,  ..., False, False, False], device='cuda:0'),
        tensor([False, False,  True,  ..., False,  True,  True], device='cuda:0'),
        tensor([ True,  True, False,  ...,  True, False, False], device='cuda:0')]
        '''
        return {'train':train_mask, 'valid': val_mask, 'test': test_mask}
        
        # return [train_mask, val_mask, test_mask]

class MultiHyperaphDatasets:
    def __init__(self, x, y, num_hypergraphs, num_features, num_classes): 

        self.x = x 
        self.y = y
        self.num_hypergraphs = num_hypergraphs
        self.num_features = num_features
        self.num_classes = num_classes

    def __len__(self):
        return self.num_hypergraphs 

    def __getitem__(self, idx):

        return self.x[idx]





# single-hypergraph 
_single_datasets_= ['cora', 'citeseer', 'pubmed',
                    'coauthor_cora', 'coauthor_dblp',
                    '20newsW100', 'ModelNet40','NTU2012', 
                    'Mushroom','zoo',
                    'yelp','walmart-trips-100','house-committees-100','walmart-trips','house-committees',
                    'actor','amazon','pokec','twitch',
                    'bail','credit','german',
                    'amazon_review','magpm','magpm_mini',
                    'trivago','ogbn_mag']

# multi-hypergraph
_multi_datasets_=["RHG_3", "RHG_10", "RHG_table", "RHG_pyramid",
                  "IMDB_dir_form", "IMDB_dir_genre","IMDB_wri_form", "IMDB_wri_genre",
                  "stream_player","twitter_friend"]

# fair-hypergraph
_fair_datasets_ = ['credit','german','bail'] 


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch_sparse
from torch_scatter import scatter, scatter_mean, scatter_add
from torch_geometric.nn.conv import MessagePassing
from dhgbench.lib_models.HNN.mlp import MLP
from dhgbench.lib_models.HNN.utils import zeros,glorot

def batched_sym_matrix_pow(matrices: torch.Tensor, p: float) -> torch.Tensor:
        r"""
        Power of a matrix using Eigen Decomposition.
        Args:
            matrices: A batch of matrices.
            p: Power.
            positive_definite: If positive definite
        Returns:
            Power of each matrix in the batch.
        """
        # vals, vecs = torch.linalg.eigh(matrices)
        # SVD is much faster than  vals, vecs = torch.linalg.eigh(matrices) for large batches.
        vecs, vals, _ = torch.linalg.svd(matrices)
        good = vals > vals.max(-1, True).values * vals.size(-1) * torch.finfo(vals.dtype).eps
        vals = vals.pow(p).where(good, torch.zeros((), device=matrices.device, dtype=matrices.dtype))
        matrix_power = (vecs * vals.unsqueeze(-2)) @ torch.transpose(vecs, -2, -1)
        return matrix_power

def sparse_diagonal(diag, shape):
    r,c = shape
    assert r == c
    indexes = torch.arange(r).to(diag.device)
    indexes = torch.stack([indexes, indexes], dim=0)
    return torch.sparse.FloatTensor(indexes, diag)

def generate_indices_general(indexes, d):
    d_range = torch.arange(d)
    d_range_edges = d_range.repeat(d).view(-1,1) #0,1..d,0,1..d..   d*d elems
    d_range_nodes = d_range.repeat_interleave(d).view(-1,1) #0,0..0,1,1..1..d,d..d  d*d elems
    indexes = indexes.unsqueeze(1) 

    large_indexes_0 = d * indexes[0] + d_range_nodes
    large_indexes_0 = large_indexes_0.permute((1,0)).reshape(1,-1)
    large_indexes_1 = d * indexes[1] + d_range_edges
    large_indexes_1 = large_indexes_1.permute((1,0)).reshape(1,-1)
    large_indexes = torch.concat((large_indexes_0, large_indexes_1), 0)

    return large_indexes

# helper functions to predict sigma(MLP(x_v || h_e)) varying how thw attributes for hyperedge are computed
def predict_blocks(x, e, hyperedge_index, sheaf_lin, args):
    # e_j = avg(x_v)
    row, col = hyperedge_index
    xs = torch.index_select(x, dim=0, index=row)
    es = torch.index_select(e, dim=0, index=col)

    # sigma(MLP(x_v || h_e))
    h_sheaf = torch.cat((xs,es), dim=-1) #sparse version of an NxEx2f tensor
    h_sheaf = sheaf_lin(h_sheaf)  #sparse version of an NxExd tensor
    if args.sheaf_act == 'sigmoid':
        h_sheaf = F.sigmoid(h_sheaf) # output d numbers for every entry in the incidence matrix
    elif args.sheaf_act == 'tanh':
        h_sheaf = F.tanh(h_sheaf) # output d numbers for every entry in the incidence matrix
    return h_sheaf

def predict_blocks_var2(x, hyperedge_index, sheaf_lin, args):
    # e_j = avg(h_v)
    row, col = hyperedge_index
    e = scatter_mean(x[row],col, dim=0)
    
    xs = torch.index_select(x, dim=0, index=row)
    es= torch.index_select(e, dim=0, index=col)

    # sigma(MLP(x_v || h_e))
    h_sheaf = torch.cat((xs,es), dim=-1) #sparse version of an NxEx2f tensor
    h_sheaf = sheaf_lin(h_sheaf)  #sparse version of an NxExd tensor
    if args.sheaf_act == 'sigmoid':
        h_sheaf = F.sigmoid(h_sheaf) # output d numbers for every entry in the incidence matrix
    elif args.sheaf_act == 'tanh':
        h_sheaf = F.tanh(h_sheaf) # output d numbers for every entry in the incidence matrix
    
    return h_sheaf

def predict_blocks_var3(x, hyperedge_index, sheaf_lin, sheaf_lin2, args):
    # universal approx according to  Equivariant Hypergraph Diffusion Neural Operators
    # # e_j = sum(φ(x_v))

    row, col = hyperedge_index
    xs = torch.index_select(x, dim=0, index=row)

    #φ(x_v)
    x_e = sheaf_lin2(x)
    #sum(φ(x_v)
    e = scatter_add(x_e[row],col, dim=0)  
    es= torch.index_select(e, dim=0, index=col)

    # sigma(MLP(x_v || h_e))
    h_sheaf = torch.cat((xs,es), dim=-1) #sparse v ersion of an NxEx2f tensor
    h_sheaf = sheaf_lin(h_sheaf)  #sparse version of an NxExd tensor
    if args.sheaf_act == 'sigmoid':
        h_sheaf = F.sigmoid(h_sheaf) # output d numbers for every entry in the incidence matrix
    elif args.sheaf_act == 'tanh':
        h_sheaf = F.tanh(h_sheaf) # output d numbers for every entry in the incidence matrix
    
    return h_sheaf

def predict_blocks_cp_decomp(x, hyperedge_index, cp_W, cp_V, sheaf_lin, args):
    row, col = hyperedge_index
    xs = torch.index_select(x, dim=0, index=row)

    xs_ones = torch.cat((xs, torch.ones(xs.shape[0],1).to(xs.device)), dim=-1) #nnz x f+1
    xs_ones_proj = torch.tanh(cp_W(xs_ones)) #nnz x r
    xs_prod =  scatter(xs_ones_proj, col, dim=0, reduce="mul") #edges x r
    e = torch.relu(cp_V(xs_prod))#edges x f
    e = e + torch.relu(scatter_add(x[row],col, dim=0))
    es= torch.index_select(e, dim=0, index=col)

    # sigma(MLP(x_v || h_e))
    h_sheaf = torch.cat((xs,es), dim=-1) #sparse version of an NxEx2f tensor
    h_sheaf = sheaf_lin(h_sheaf)  #sparse version of an NxExd tensor
    if args.sheaf_act == 'sigmoid':
        h_sheaf = F.sigmoid(h_sheaf) # output d numbers for every entry in the incidence matrix
    elif args.sheaf_act == 'tanh':
        h_sheaf = F.tanh(h_sheaf) # output d numbers for every entry in the incidence matrix
    return h_sheaf

# Build the restriction maps for the Diagonal Case
class SheafBuilderDiag(nn.Module):
    def __init__(self, args):
        super(SheafBuilderDiag, self).__init__()
        self.args = args
        self.prediction_type = args.sheaf_pred_block # pick the way hyperedge feartures are computed
        self.sheaf_dropout = args.sheaf_dropout
        self.special_head = args.sheaf_special_head # add a head having just 1 on the diagonal. this should be similar to the normal hypergraph conv
        self.d = args.stalk_dim # stalk dinension
        self.MLP_hidden = args.MLP_hidden
        self.norm = args.AllSet_input_norm
        self.dropout = args.dropout
        
        self.sheaf_lin = MLP(
                    in_channels=2*self.MLP_hidden, 
                    hidden_channels=args.MLP_hidden,
                    out_channels=self.d,
                    num_layers=1,
                    dropout=0.0,
                    Normalization='ln',
                    InputNorm=self.norm)

        if self.prediction_type == 'MLP_var3':
            self.sheaf_lin2 = MLP(
                        in_channels=self.MLP_hidden, 
                        hidden_channels=args.MLP_hidden,
                        out_channels=args.MLP_hidden,
                        num_layers=1,
                        dropout=0.0,
                        Normalization='ln',
                        InputNorm=self.norm)
        elif self.prediction_type == 'cp_decomp':
            self.cp_W = MLP(
                        in_channels=self.MLP_hidden+1, 
                        hidden_channels=args.MLP_hidden,
                        out_channels=args.MLP_hidden,
                        num_layers=1,
                        dropout=0.0,
                        Normalization='ln',
                        InputNorm=self.norm)
            self.cp_V = MLP(
                        in_channels=args.MLP_hidden, 
                        hidden_channels=args.MLP_hidden,
                        out_channels=args.MLP_hidden,
                        num_layers=1,
                        dropout=0.0,
                        Normalization='ln',
                        InputNorm=self.MLP_hidden)        

    def reset_parameters(self):
        if self.prediction_type == 'MLP_var3':
            self.sheaf_lin.reset_parameters()
            self.sheaf_lin2.reset_parameters()
        elif self.prediction_type == 'cp_decomp':
            self.sheaf_lin.reset_parameters()
            self.cp_W.reset_parameters()
            self.cp_V.reset_parameters()
        else:
            self.sheaf_lin.reset_parameters()
        
    # this is exclusively for diagonal sheaf
    def forward(self, x, e, hyperedge_index):
        """ tmp
        x: Nd x f -> N x f
        e: Ed x f -> E x f
        -> (concat) N x E x (d+1)F -> (linear project) N x E x d (the elements on the diagonal of each dxd block)
        -> (reshape) (Nd x Ed) with NxE diagonal blocks of dimension dxd

        """
        num_nodes = x.shape[0] // self.d
        num_edges = hyperedge_index[1].max().item() + 1
        x = x.view(num_nodes, self.d, x.shape[-1]).mean(1) # N x d x f -> N x f
        e = e.view(num_edges, self.d, e.shape[-1]).mean(1) # # x d x f -> E x f

        #predict (_ x d) elements
        if self.prediction_type == 'MLP_var1':
            h_sheaf = predict_blocks(x, e, hyperedge_index, self.sheaf_lin, self.args)
        elif self.prediction_type == 'MLP_var2':
            h_sheaf = predict_blocks_var2(x, hyperedge_index, self.sheaf_lin, self.args)
        elif self.prediction_type == 'MLP_var3':
            h_sheaf = predict_blocks_var3(x, hyperedge_index, self.sheaf_lin,  self.sheaf_lin2, self.args)
        elif self.prediction_type == 'cp_decomp':
            h_sheaf = predict_blocks_cp_decomp(x, hyperedge_index, self.cp_W, self.cp_V, self.sheaf_lin, self.args)
        
        if self.sheaf_dropout:
            h_sheaf = F.dropout(h_sheaf, p=self.dropout, training=self.training)
        
        if self.special_head:
            new_head_mask = [1]*(self.d-1) + [0]
            new_head = [0]*(self.d-1) + [1]
            h_sheaf = h_sheaf * torch.tensor(new_head_mask, device=x.device) + torch.tensor(new_head, device=x.device)
        
        self.h_sheaf = h_sheaf #this is stored in self for testing purpose
        h_sheaf_attributes = h_sheaf.reshape(-1) #(d*K)

        # from a d-dim tensor assoc to every entrence in edge_index
        # create a sparse incidence Nd x Ed

        # We need to modify indices from the NxE matrix 
        # to correspond to the large Nd x Ed matrix, but restrict only on the element of the diagonal of each block
        # indices: scalar [i,j] -> block dxd with indices [d*i, d*i+1.. d*i+d-1; d*j, d*j+1 .. d*j+d-1]
        # attributes: reshape h_sheaf

        d_range = torch.arange(self.d, device=x.device).view(1,-1,1).repeat(2,1,1) #2xdx1
        hyperedge_index = hyperedge_index.unsqueeze(1) #2x1xK
        hyperedge_index = self.d * hyperedge_index + d_range #2xdxK
        hyperedge_index = hyperedge_index.permute(0,2,1).reshape(2,-1) #2x(d*K)
        h_sheaf_index = hyperedge_index

        #the resulting (index, values) pair correspond to the diagonal of each block sub-matrix
        return h_sheaf_index, h_sheaf_attributes

def normalisation_matrices(x, hyperedge_index, alpha, num_nodes, num_edges, d, norm_type='degree_norm'):
    #this will return either D^-1/B^-1 or D^(-1/2)/B^-1
    if norm_type == 'degree_norm':
        #return D_inv and B_inv used to normalised the laplacian (/propagation)
        #normalise using node/hyperedge degrees D_e and D_v in the paper
        D = scatter_add(x.new_ones(hyperedge_index.size(1)), hyperedge_index[0], dim=0, dim_size=num_nodes*d) 
        D = 1.0 / D
        D[D == float("inf")] = 0

        B = scatter_add(x.new_ones(hyperedge_index.size(1)), hyperedge_index[1], dim=0, dim_size=num_edges*d)
        B = 1.0 / B
        B[B == float("inf")] = 0
        return D, B

    elif norm_type == 'sym_degree_norm':
        #normalise using node/hyperedge degrees D_e and D_v in the paper
        D = scatter_add(x.new_ones(hyperedge_index.size(1)), hyperedge_index[0], dim=0, dim_size=num_nodes*d) 
        D = D ** (-0.5)
        D[D == float("inf")] = 0

        B = scatter_add(x.new_ones(hyperedge_index.size(1)), hyperedge_index[1], dim=0, dim_size=num_edges*d)
        B = 1.0 / B
        B[B == float("inf")] = 0
        return D, B

    elif norm_type == 'block_norm':
        #normalise using diag(HHT) and deg_e <- this take into account the values predicted in H as oposed to 0/1 as in the degree
        # this way of computing the normalisation tensor is only valid for diagonal sheaf
        D = scatter_add(alpha*alpha, hyperedge_index[0], dim=0, dim_size=num_nodes*d)
        D = 1.0 / D #can compute inverse like this because the matrix is diagonal
        D[D == float("inf")] = 0

        B = scatter_add(x.new_ones(hyperedge_index.size(1)), hyperedge_index[1], dim=0, dim_size=num_edges*d)
        B = 1.0 / B
        B[B == float("inf")] = 0
        return D, B
    
    elif norm_type == 'sym_block_norm':
        #normalise using diag(HHT) and deg_e <- this take into account the values predicted in H as oposed to 0/1 as in the degree
        # this way of computing the normalisation tensor is only valid for diagonal sheaf
        D = scatter_add(alpha*alpha, hyperedge_index[0], dim=0, dim_size=num_nodes*d)
        D = D ** (-0.5) #can compute inverse like this because the matrix is diagonal
        D[D == float("inf")] = 0

        B = scatter_add(x.new_ones(hyperedge_index.size(1)), hyperedge_index[1], dim=0, dim_size=num_edges*d)
        B = 1.0 / B
        B[B == float("inf")] = 0
        return D, B

# One layer of Sheaf Diffusion with diagonal Laplacian Y = (I-D^-1/2LD^-1) with L normalised with B^-1
class HyperDiffusionDiagSheafConv(MessagePassing):
    r"""
    
    """
    def __init__(self, in_channels, out_channels, d, device, dropout=0, bias=True, norm_type='degree_norm', 
                left_proj=None, norm=None, residual = False, clear_flag=False,
                 **kwargs):
        kwargs.setdefault('aggr', 'add')
        super().__init__(flow='source_to_target', node_dim=0, **kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.d = d
        self.norm_type = norm_type
        self.left_proj = left_proj
        self.norm = norm
        self.residual = residual

        if self.left_proj:
            self.lin_left_proj = MLP(in_channels=d, 
                        hidden_channels=d,
                        out_channels=d,
                        num_layers=1,
                        dropout=0.0,
                        Normalization='ln',
                        InputNorm=self.norm)
                        
        self.lin = MLP(in_channels=in_channels, 
                        hidden_channels=out_channels,
                        out_channels=out_channels,
                        num_layers=1,
                        dropout=0.0,
                        Normalization='ln',
                        InputNorm=self.norm)      
        
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)

        self.device = device
        
        self.clear_flag=clear_flag

        self.I_mask = None
        self.Id = None

        self.reset_parameters()

    #to allow multiple runs reset all parameters used
    def reset_parameters(self):
        if self.left_proj:
            self.lin_left_proj.reset_parameters()
        self.lin.reset_parameters()
        zeros(self.bias)

    def forward(self, x , hyperedge_index,
                alpha, 
                num_nodes,
                num_edges):
        r"""
        Args:
            x (Tensor): Node feature matrix {Nd x F}`.
            hyperedge_index (LongTensor): The hyperedge indices, *i.e.*
                the sparse incidence matrix Nd x Md} from nodes to edges.
            alpha (Tensor, optional): restriction maps
        """ 
        if self.left_proj:
            x = x.t().reshape(-1, self.d)
            x = self.lin_left_proj(x)
            x = x.reshape(-1,num_nodes * self.d).t()
        x = self.lin(x)
        data_x = x

        #depending on norm_type D^-1 or D^-1/2
        D_inv, B_inv = normalisation_matrices(x, hyperedge_index, alpha, num_nodes, num_edges, self.d, self.norm_type)

        if self.norm_type in ['sym_degree_norm', 'sym_block_norm']:
            # compute D^(-1/2) @ X
            x = D_inv.unsqueeze(-1) * x

        H = torch.sparse.FloatTensor(hyperedge_index, alpha, size=(num_nodes*self.d, num_edges*self.d))
        H_t = torch.sparse.FloatTensor(hyperedge_index.flip([0]), alpha, size=(num_edges*self.d, num_nodes*self.d))

        #this is because spdiags does not support gpu
        B_inv =  sparse_diagonal(B_inv, shape = (num_edges*self.d, num_edges*self.d))
        D_inv = sparse_diagonal(D_inv, shape = (num_nodes*self.d, num_nodes*self.d))

        B_inv = B_inv.coalesce()
        H_t = H_t.coalesce()
        H = H.coalesce()
        D_inv = D_inv.coalesce()

        minus_L = torch_sparse.spspmm(B_inv.indices(), B_inv.values(), H_t.indices(), H_t.values(), B_inv.shape[0],B_inv.shape[1], H_t.shape[1])
        minus_L = torch_sparse.spspmm(H.indices(), H.values(), minus_L[0], minus_L[1], H.shape[0],H.shape[1], H_t.shape[1])
        minus_L = torch_sparse.spspmm(D_inv.indices(), D_inv.values(), minus_L[0], minus_L[1], D_inv.shape[0],D_inv.shape[1], H_t.shape[1])
        minus_L = torch.sparse_coo_tensor(minus_L[0], minus_L[1], size=(num_nodes*self.d, num_nodes*self.d)).to(self.device)

        #negate the diagonal blocks and add eye matrix
        if self.I_mask is None or self.clear_flag: #prepare these in advance
            I_mask_indices = torch.stack([torch.arange(num_nodes), torch.arange(num_nodes)], dim=0)
            I_mask_indices = generate_indices_general(I_mask_indices, self.d)
            I_mask_values = torch.ones((I_mask_indices.shape[1]))
            self.I_mask = torch.sparse_coo_tensor(I_mask_indices, I_mask_values).to(self.device)
            self.Id = sparse_diagonal(torch.ones(num_nodes*self.d), shape = (num_nodes*self.d, num_nodes * self.d)).to(self.device)

        minus_L = minus_L.coalesce()
        #this help us changing the sign of the elements in the block diagonal
        #with an efficient lower=memory mask 
        minus_L = torch.sparse_coo_tensor(minus_L.indices(), minus_L.values(), minus_L.size())
        minus_L = minus_L - 2 * minus_L.mul(self.I_mask)
        minus_L = self.Id + minus_L

        minus_L = minus_L.coalesce()
        out = torch_sparse.spmm(minus_L.indices(), minus_L.values(), minus_L.shape[0], minus_L.shape[1], x)
        if self.bias is not None:
            out = out + self.bias
        if self.residual:
            out = out + data_x
        return out
    
class SheafHyperGNN(nn.Module):
    """
        This is a Hypergraph Sheaf Model with 
        the dxd blocks in H_BIG associated to each pair (node, hyperedge)
        being **diagonal**
    """
    def __init__(self, num_features, num_targets, args, sheaf_type='SheafHyperGNNDiag'):
        super(SheafHyperGNN, self).__init__()

        self.num_layers = args.All_num_layers
        self.dropout = args.dropout  # Note that default is 0.6
        self.num_features = num_features
        self.num_targets = num_targets
        self.MLP_hidden = args.MLP_hidden 
        self.d = args.stalk_dim  # dimension of the stalks
        self.init_hedge = args.init_hedge # how to initialise hyperedge attributes: avg or rand
        self.norm_type = args.sheaf_normtype #type of laplacian normalisation degree_norm or block_norm
        self.act = args.sheaf_act # type of nonlinearity used when predicting the dxd blocks
        self.left_proj = args.sheaf_left_proj # multiply with (I x W_1) to the left
        self.args = args
        self.norm = args.AllSet_input_norm
        self.dynamic_sheaf = args.dynamic_sheaf # if True, the sheaf changes from one layer to another
        self.residual = args.residual_sheaf

        self.clear_flag = True if args.task_type == 'hg_cls' else False # 如果是Hypergraph Classification Task，由于要考虑多图情况，需要每轮清空部分参数
           
        self.hyperedge_attr = None
        
        if args.device in [0, 1]:
            self.device = torch.device('cuda:'+str(args.device)
                              if torch.cuda.is_available() else 'cpu')
        else:
            self.device = torch.device(args.device)
        
        self.lin = MLP(in_channels=self.num_features, 
                        hidden_channels=args.MLP_hidden,
                        out_channels=self.MLP_hidden*self.d,
                        num_layers=1,
                        dropout=0.0,
                        Normalization='ln',
                        InputNorm=False)
        
        # define the model and sheaf generator according to the type of sheaf wanted
        # The diuffusion does not change, however tha implementation for diag and ortho is more efficient
        if sheaf_type == 'SheafHyperGNNDiag':
            ModelSheaf, SheafConv = SheafBuilderDiag, HyperDiffusionDiagSheafConv
        else:
            raise NotImplementedError
        
        self.convs = nn.ModuleList()
        # Sheaf Diffusion layers
        self.convs.append(SheafConv(self.MLP_hidden, self.MLP_hidden, d=self.d, device=self.device, 
                                        norm_type=self.norm_type, left_proj=self.left_proj, 
                                        norm=self.norm, residual=self.residual, clear_flag=self.clear_flag))
                                        
        # Model to generate the reduction maps
        self.sheaf_builder = nn.ModuleList()
        self.sheaf_builder.append(ModelSheaf(args))

        for _ in range(self.num_layers-1):
            # Sheaf Diffusion layers
            self.convs.append(SheafConv(self.MLP_hidden, self.MLP_hidden, d=self.d, device=self.device, 
                                        norm_type=self.norm_type, left_proj=self.left_proj, 
                                        norm=self.norm, residual=self.residual, clear_flag=self.clear_flag))
            # Model to generate the reduction maps if the sheaf changes from one layer to another
            if self.dynamic_sheaf:
                self.sheaf_builder.append(ModelSheaf(args))
                
        self.lin2 = nn.Linear(self.MLP_hidden*self.d, self.num_targets, bias=False)

    def reset_parameters(self):
        for conv in self.convs:
            conv.reset_parameters()
        for sheaf_builder in self.sheaf_builder:
            sheaf_builder.reset_parameters()

        self.lin.reset_parameters()
        self.lin2.reset_parameters()    

    def init_hyperedge_attr(self, type, num_edges=None, x=None, hyperedge_index=None):
        #initialize hyperedge attributes either random or as the average of the nodes
        if type == 'rand':
            hyperedge_attr = torch.randn((num_edges, self.num_features)).to(self.device)
        elif type == 'avg':
            hyperedge_attr = scatter_mean(x[hyperedge_index[0]],hyperedge_index[1], dim=0)
        else:
            hyperedge_attr = None
        return hyperedge_attr

    def forward(self, data):
        x = data.x
        edge_index = data.hyperedge_index
        num_nodes = data.x.shape[0] #data.edge_index[0].max().item() + 1
        num_edges = data.hyperedge_index[1].max().item() + 1

        #if we are at the first epoch, initialise the attribute, otherwise use the previous ones
        if self.hyperedge_attr is None:
            self.hyperedge_attr = self.init_hyperedge_attr(self.init_hedge, num_edges=num_edges, x=x, hyperedge_index=edge_index)

        # expand the input N x num_features -> Nd x num_features such that we can apply the propagation
        x = self.lin(x)
        x = x.view((x.shape[0]*self.d, self.MLP_hidden)) # (N * d) x num_features

        hyperedge_attr = self.lin(self.hyperedge_attr)
        hyperedge_attr = hyperedge_attr.view((hyperedge_attr.shape[0]*self.d, self.MLP_hidden))

        for i, conv in enumerate(self.convs[:-1]):
            # infer the sheaf as a sparse incidence matrix Nd x Ed, with each block being diagonal
            if i == 0 or self.dynamic_sheaf:
                h_sheaf_index, h_sheaf_attributes = self.sheaf_builder[i](x, hyperedge_attr, edge_index)
            # Sheaf Laplacian Diffusion
            x = F.elu(conv(x, hyperedge_index=h_sheaf_index, alpha=h_sheaf_attributes, num_nodes=num_nodes, num_edges=num_edges))
            x = F.dropout(x, p=self.dropout, training=self.training)

        #infer the sheaf as a sparse incidence matrix Nd x Ed, with each block being diagonal
        if len(self.convs) == 1 or self.dynamic_sheaf:
            h_sheaf_index, h_sheaf_attributes = self.sheaf_builder[-1](x, hyperedge_attr, edge_index) 
        # Sheaf Laplacian Diffusion
        x = self.convs[-1](x,  hyperedge_index=h_sheaf_index, alpha=h_sheaf_attributes, num_nodes=num_nodes, num_edges=num_edges)
        x = x.view(num_nodes, -1) # Nd x out_channels -> Nx(d*out_channels)

        x = self.lin2(x) # Nx(d*out_channels)-> N x num_classes 
        
        return x, None


import torch.nn as nn, torch.nn.functional as F
from torch_geometric.nn import GCNConv
import torch_scatter

class LEGCN(nn.Module):

    def __init__(self, num_features, num_targets, args):
        """
        d: initial node-feature dimension
        h: number of hidden units
        c: number of classes
        """
        super(LEGCN, self).__init__()
        self.hidden_dim = args.MLP_hidden
        self.num_layers = args.All_num_layers
        self.dropout=args.dropout
        self.convs = nn.ModuleList([])
        if self.num_layers == 1:
            self.convs.append(GCNConv(in_channels=num_features, out_channels=num_targets))
        else:
            self.convs.append(GCNConv(in_channels=num_features, out_channels=self.hidden_dim))
            for l in range(self.num_layers - 2):
                self.convs.append(GCNConv(in_channels=self.hidden_dim, out_channels=self.hidden_dim))
            self.convs.append(GCNConv(in_channels=self.hidden_dim, out_channels=num_targets))

    def reset_parameters(self):
        for layer in self.convs:
            layer.reset_parameters()

    def forward(self, data):
        """
        an l-layer GCN
        """
        le_adj = data.le_adj
        edge_index = data.hyperedge_index
        x = data.x[edge_index[0]]
        for conv in self.convs[:-1]:
            x = F.relu(conv(x, le_adj))
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.convs[-1](x, le_adj)
        x = torch_scatter.scatter(x, edge_index[0], dim=0, reduce='mean')
        return x,None 


#!/usr/bin/env python
# coding: utf-8

# In[45]:

import torch
import pickle
import os
import os.path as osp
from torch_geometric.data import InMemoryDataset
from dhgbench.lib_dataset.load_other_datasets import * 

class dataset_Hypergraph(InMemoryDataset):
    
    def __init__(self, root = '../data/pyg_data/hypergraph_dataset_updated/', name = None, 
                 p2raw = None,
                 train_percent = 0.01,
                 feature_noise = None,
                 transform=None, pre_transform=None):
        
        existing_dataset = ['cora', 'citeseer', 'pubmed',
                            'coauthor_cora', 'coauthor_dblp',
                            '20newsW100', 'ModelNet40', 'NTU2012',
                            'zoo', 'Mushroom', 
                            'yelp', 'amazon-reviews', 'walmart-trips', 'house-committees',
                            'walmart-trips-100', 'house-committees-100',
                            'bail','credit','german',
                            'actor','amazon','pokec','twitch',
                            'amazon_review','magpm','magpm_mini','trivago','ogbn_mag']
        
        if name not in existing_dataset:
            raise ValueError(f'name of hypergraph dataset must be one of: {existing_dataset}')
        else:
            self.name = name
        
        self.feature_noise = feature_noise
        self._train_percent = train_percent

        if (p2raw is not None) and osp.isdir(p2raw):
            self.p2raw = p2raw
        elif p2raw is None:
            self.p2raw = None
        elif not osp.isdir(p2raw):
            raise ValueError(f'path to raw hypergraph dataset "{p2raw}" does not exist!')
        
        if not osp.isdir(root):
            os.makedirs(root)
            
        self.root = root
        self.myraw_dir = osp.join(root, self.name, 'raw')
        self.myprocessed_dir = osp.join(root, self.name, 'processed')
        
        super(dataset_Hypergraph, self).__init__(osp.join(root, name), transform, pre_transform)
        
        self.data, self.slices = torch.load(self.processed_paths[0], weights_only=False)
        self.train_percent = self.data.train_percent

    @property
    def raw_file_names(self):
        if self.feature_noise is not None:
            file_names = [f'{self.name}_noise_{self.feature_noise}']
        else:
            file_names = [self.name]
        return file_names

    @property
    def processed_file_names(self):
        if self.feature_noise is not None:
            file_names = [f'data_noise_{self.feature_noise}.pt']
        else:
            file_names = ['data.pt']
        return file_names

    @property
    def num_features(self):
        return self.data.num_node_features

    def download(self):
        for name in self.raw_file_names:
            p2f = osp.join(self.myraw_dir, name)
            if not osp.isfile(p2f):
                # file not exist, so we create it and save it there.
                print(p2f)
                print(self.p2raw)
                print(self.name)

                if self.name in ['cora', 'citeseer', 'pubmed']:
                    tmp_data = load_dataset(path = self.p2raw,
                            dataset = self.name, 
                            train_percent = self._train_percent)

                elif self.name in ['coauthor_cora', 'coauthor_dblp']:
                    assert 'coauthorship' in self.p2raw
                    dataset_name = self.name.split('_')[-1]
                    tmp_data = load_dataset(path = self.p2raw,
                            dataset = dataset_name,
                            train_percent = self._train_percent)
                    
                elif self.name in ['bail','credit','german']:
                    tmp_data = load_dataset(path = self.p2raw,
                            dataset = self.name, 
                            train_percent = self._train_percent)
                    
                elif self.name in ['actor','amazon','pokec','twitch']:
                    tmp_data = load_dataset(path = self.p2raw,
                            dataset = self.name, 
                            train_percent = self._train_percent)

                elif self.name in ['amazon-reviews', 'walmart-trips', 'house-committees']:
                    if self.feature_noise is None:
                        raise ValueError(f'for cornell datasets, feature noise cannot be {self.feature_noise}')
                    tmp_data = load_cornell_dataset(path = self.p2raw,
                        dataset = self.name,
                        feature_noise = self.feature_noise,
                        train_percent = self._train_percent)
                    
                elif self.name in ['walmart-trips-100', 'house-committees-100']:
                    if self.feature_noise is None:
                        raise ValueError(f'for cornell datasets, feature noise cannot be {self.feature_noise}')
                    feature_dim = int(self.name.split('-')[-1])
                    tmp_name = '-'.join(self.name.split('-')[:-1])
                    tmp_data = load_cornell_dataset(path = self.p2raw,
                        dataset = tmp_name,
                        feature_dim = feature_dim,
                        feature_noise = self.feature_noise,
                        train_percent = self._train_percent)

                elif self.name == 'yelp':
                    tmp_data = load_yelp_dataset(path = self.p2raw,
                            dataset = self.name,
                            train_percent = self._train_percent)

                elif self.name in ['amazon_review','magpm','magpm_mini','trivago','ogbn_mag']:
                    tmp_data = load_large_dataset(path = self.p2raw,
                            dataset = self.name,
                            train_percent = self._train_percent)
                
                else:
                    tmp_data = load_LE_dataset(path = self.p2raw, 
                                              dataset = self.name,
                                              train_percent= self._train_percent)
                    
                _ = save_data_to_pickle(tmp_data, 
                                          p2root = self.myraw_dir,
                                          file_name = self.raw_file_names[0])
            else:
                raise ValueError('Undefined data')

    def process(self):
        p2f = osp.join(self.myraw_dir, self.raw_file_names[0])
        with open(p2f, 'rb') as f:
            data = pickle.load(f)
        data = data if self.pre_transform is None else self.pre_transform(data)
        torch.save(self.collate([data]), self.processed_paths[0])

    def __repr__(self):
        return '{}()'.format(self.name)

def save_data_to_pickle(data, p2root = '../data/', file_name = None):
    '''
    if file name not specified, use time stamp.
    '''
#     now = datetime.now()
#     surfix = now.strftime('%b_%d_%Y-%H:%M')
    surfix = 'star_expansion_dataset'
    if file_name is None:
        tmp_data_name = '_'.join(['Hypergraph', surfix])
    else:
        tmp_data_name = file_name
    p2he_StarExpan = osp.join(p2root, tmp_data_name)
    if not osp.isdir(p2root):
        os.makedirs(p2root)
    with open(p2he_StarExpan, 'bw') as f:
        pickle.dump(data, f)
    return p2he_StarExpan



import copy
import torch
from torch import Tensor

def perturbation(data,mode,p,masks=None):

    perturbed_data = copy.deepcopy(data)
    
    if mode == 'add_incidence':
        pt_hyperedge_index = add_incidence(data.hyperedge_index,p)
        perturbed_data.hyperedge_index = pt_hyperedge_index
    elif mode == 'drop_incidence':
        pt_hyperedge_index = drop_incidence(data.hyperedge_index,p,data.num_nodes,data.num_hyperedges)
        perturbed_data.hyperedge_index = pt_hyperedge_index
    elif mode == 'spar_feat':
        pt_feat = drop_features(data.x,p)
        perturbed_data.x = pt_feat
    elif mode == 'noise_feat':
        pt_feat = add_gaussian_noise(data.x,p)
        perturbed_data.x = pt_feat
    elif mode == 'spar_label':
        new_train_mask = mask_label(masks['train'], p=p)
        masks['train'] = new_train_mask
        return masks 
    elif mode == 'flip_label':
        perturbed_labels = perturb_train_labels(data.y,masks['train'],data.num_classes,p)
        perturbed_data.y = perturbed_labels
    else:
        raise NotImplementedError('Not Implemneted Perturbation')
    
    return perturbed_data

def drop_features(x: Tensor, p: float):

    drop_mask = torch.empty((x.size(1), ), dtype=torch.float32, device=x.device).uniform_(0, 1) < p
    x = x.clone()
    x[:, drop_mask] = 0
    return x

def add_gaussian_noise(x: Tensor, lam: float):

    r = x.max(dim=1, keepdim=True).values.mean()
    noise = torch.randn_like(x) * lam * r

    return x+noise

def filter_incidence(row: Tensor, col: Tensor, hyperedge_attr, mask: Tensor):
    return row[mask], col[mask], None if hyperedge_attr is None else hyperedge_attr[mask]

def drop_incidence(hyperedge_index: Tensor, p: float = 0.2, num_nodes: int = 0, num_edges: int = 0) -> Tensor:
    if p == 0.0:
        return hyperedge_index
    
    row, col = hyperedge_index
    mask = torch.rand(row.size(0), device=hyperedge_index.device) >= p
    
    row, col, _ = filter_incidence(row, col, None, mask)
    hyperedge_index = torch.stack([row, col], dim=0)

    # padding operation
    padding_edge = torch.tensor([[num_nodes-1], [num_edges-1]],device=hyperedge_index.device)
    hyperedge_index = torch.cat([hyperedge_index, padding_edge], dim=1)

    return hyperedge_index

def add_incidence(hyperedge_index: Tensor, p: float = 0.2):

    row, col = hyperedge_index
    num_nodes, num_edges= row.max()+1,col.max()+1

    num_add = int(row.size(0) * p) 
    
    if num_add > 0:
        existing_edges = set(zip(row.tolist(), col.tolist()))
        new_edges = set()
        
        while len(new_edges) < num_add:
            new_edge = (torch.randint(0, num_nodes, (1,)).item(), 
                        torch.randint(0, num_edges, (1,)).item())
            if new_edge not in existing_edges:
                new_edges.add(new_edge)
        
        new_rows, new_cols = zip(*new_edges)
        row = torch.cat([row, torch.tensor(new_rows, device=row.device)], dim=0)
        col = torch.cat([col, torch.tensor(new_cols, device=col.device)], dim=0)
    
    return torch.stack([row, col], dim=0)

def perturb_train_labels(labels: torch.Tensor, train_mask: torch.Tensor, num_classes: int, p: float = 0.1):

    perturbed_labels = labels.clone()
    
    train_indices = torch.where(train_mask)[0]
    num_perturb = int(len(train_indices) * p) 
    
    if num_perturb > 0:
        perturb_indices = train_indices[torch.randperm(len(train_indices))[:num_perturb]]
        
        random_labels = torch.randint(0, num_classes, (num_perturb,), device=labels.device)
        mask_diff = random_labels == labels[perturb_indices]
        while mask_diff.any(): 
            random_labels[mask_diff] = torch.randint(0, num_classes, (mask_diff.sum(),), device=labels.device)
            mask_diff = random_labels == labels[perturb_indices]
        
        perturbed_labels[perturb_indices] = random_labels
    
    return perturbed_labels

def mask_label(mask: torch.Tensor, p: float):

    assert mask.dtype == torch.bool

    true_indices = torch.nonzero(mask, as_tuple=False).squeeze()

    num_to_flip = int(len(true_indices) * p)

    if num_to_flip == 0:
        return mask.clone()  

    perm = torch.randperm(len(true_indices), device=mask.device)
    flip_indices = true_indices[perm[:num_to_flip]]

    new_mask = mask.clone()
    new_mask[flip_indices] = False

    return new_mask


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.utils import softmax
from torch_scatter import scatter
from torch_geometric.typing import Adj, Size, OptTensor
from dhgbench.lib_models.HNN.mlp import MLP
from dhgbench.lib_models.HNN.utils import zeros,glorot

class PMA(MessagePassing):

    _alpha: OptTensor

    def __init__(self, in_channels, hid_dim,
                 out_channels, num_layers, heads=1, concat=True,
                 negative_slope=0.2, dropout=0.0, bias=False, **kwargs):
        #         kwargs.setdefault('aggr', 'add')
        super(PMA, self).__init__(node_dim=0, **kwargs)

        self.in_channels = in_channels
        self.hidden = hid_dim // heads
        self.out_channels = out_channels
        self.heads = heads
        self.concat = concat
        self.negative_slope = negative_slope
        self.dropout = 0.
        self.aggr = 'add'

        self.lin_K = nn.Linear(in_channels, self.heads*self.hidden)
        self.lin_V = nn.Linear(in_channels, self.heads*self.hidden)
        self.att_r = nn.Parameter(torch.Tensor(
            1, heads, self.hidden))  # Seed vector
        self.rFF = MLP(in_channels=self.heads*self.hidden,
                       hidden_channels=self.heads*self.hidden,
                       out_channels=out_channels,
                       num_layers=num_layers,
                       dropout=.0, Normalization='None',)
        self.ln0 = nn.LayerNorm(self.heads*self.hidden)
        self.ln1 = nn.LayerNorm(self.heads*self.hidden)

        self.register_parameter('bias', None)

        self._alpha = None

        self.reset_parameters()

    def reset_parameters(self):
        #         glorot(self.lin_l.weight)
        glorot(self.lin_K.weight)
        glorot(self.lin_V.weight)
        self.rFF.reset_parameters()
        self.ln0.reset_parameters()
        self.ln1.reset_parameters()
#         glorot(self.att_l)
        nn.init.xavier_uniform_(self.att_r)
#         zeros(self.bias)

    def forward(self, x, edge_index: Adj,
                size: Size = None, return_attention_weights=None):
        
        H, C = self.heads, self.hidden

        x_l: OptTensor = None
        x_r: OptTensor = None
        alpha_l: OptTensor = None
        alpha_r: OptTensor = None
        if isinstance(x, Tensor):
            assert x.dim() == 2, 'Static graphs not supported in `GATConv`.'
            x_K = self.lin_K(x).view(-1, H, C)
            x_V = self.lin_V(x).view(-1, H, C)
            alpha_r = (x_K * self.att_r).sum(dim=-1)

        out = self.propagate(edge_index, x=x_V,
                             alpha=alpha_r, aggr=self.aggr)

        alpha = self._alpha
        self._alpha = None

#         Note that in the original code of GMT paper, they do not use additional W^O to combine heads.
#         This is because O = softmax(QK^T)V and V = V_in*W^V. So W^O can be effectively taken care by W^V!!!
        out += self.att_r  # This is Seed + Multihead
        # concat heads then LayerNorm. Z (rhs of Eq(7)) in GMT paper.
        out = self.ln0(out.view(-1, self.heads * self.hidden))
        # rFF and skip connection. Lhs of eq(7) in GMT paper.
        out = self.ln1(out+F.relu(self.rFF(out)))

        if isinstance(return_attention_weights, bool):
            assert alpha is not None
            if isinstance(edge_index, Tensor):
                return out, (edge_index, alpha)
            elif isinstance(edge_index, SparseTensor):
                return out, edge_index.set_value(alpha, layout='coo')
        else:
            return out

    def message(self, x_j, alpha_j,
                index, ptr,
                size_j):
        #         ipdb.set_trace()
        alpha = alpha_j
        alpha = F.leaky_relu(alpha, self.negative_slope)
        alpha = softmax(alpha, index, ptr, index.max()+1)
        self._alpha = alpha
        alpha = F.dropout(alpha, p=self.dropout, training=self.training)
        return x_j * alpha.unsqueeze(-1)

    def aggregate(self, inputs, index,
                  dim_size=None, aggr='add'):

        if aggr is None:
            raise ValueError("aggr was not passed!")
        return scatter(inputs, index, dim=self.node_dim, reduce=aggr)

    def __repr__(self):
        return '{}({}, {}, heads={})'.format(self.__class__.__name__,
                                             self.in_channels,
                                             self.out_channels, self.heads)

class HalfNLHconv(MessagePassing):
    def __init__(self,
                 in_dim,
                 hid_dim,
                 out_dim,
                 num_layers,
                 dropout,
                 Normalization='bn',
                 InputNorm=False,
                 heads=1,
                 attention=True
                 ):
        super(HalfNLHconv, self).__init__()

        self.attention = attention
        self.dropout = dropout

        if self.attention:
            self.prop = PMA(in_dim, hid_dim, out_dim, num_layers, heads=heads)
        else:
            if num_layers > 0:
                self.f_enc = MLP(in_dim, hid_dim, hid_dim, num_layers, dropout, Normalization, InputNorm)
                self.f_dec = MLP(hid_dim, hid_dim, out_dim, num_layers, dropout, Normalization, InputNorm)
            else:
                self.f_enc = nn.Identity()
                self.f_dec = nn.Identity()

    def reset_parameters(self):

        if self.attention:
            self.prop.reset_parameters()
        else:
            if not (self.f_enc.__class__.__name__ != 'Identity'):
                self.f_enc.reset_parameters()
            if not (self.f_dec.__class__.__name__ != 'Identity'):
                self.f_dec.reset_parameters()

    def forward(self, x, edge_index, norm, aggr='add'):
        
        if self.attention:
            x = self.prop(x, edge_index)
        else:
            x = F.relu(self.f_enc(x))
            x = F.dropout(x, p=self.dropout, training=self.training)
            x = self.propagate(edge_index, x=x, norm=norm, aggr=aggr)
            x = F.relu(self.f_dec(x))
            
        return x

    def message(self, x_j, norm):
        return norm.view(-1, 1) * x_j

    def aggregate(self, inputs, index,
                  dim_size=None, aggr=None):
        aggr = aggr if aggr is not None else "add"
        if aggr is None:
            raise ValueError("aggr was not passed!")
        return scatter(inputs, index, dim=self.node_dim, reduce=aggr)


class SetGNN(nn.Module):
    def __init__(self, num_features, num_targets, args, norm=None):
        super(SetGNN, self).__init__()

        self.All_num_layers = args.All_num_layers
        self.dropout = args.dropout
        self.aggr = args.aggregate
        self.NormLayer = args.normalization
        self.InputNorm = args.deepset_input_norm
        self.GPR = args.GPR
        self.LearnMask = args.LearnMask
        self.decoder_hidden = args.decoder_hidden
        self.decoder_num_layers = args.decoder_num_layers

        self.V2EConvs = nn.ModuleList()
        self.E2VConvs = nn.ModuleList()
        self.bnV2Es = nn.ModuleList()
        self.bnE2Vs = nn.ModuleList()

        if self.LearnMask:
            self.Importance = nn.Parameter(torch.ones(norm.size()))

        if self.All_num_layers == 0:
            self.decoder = MLP(in_channels=num_features,
                                  hidden_channels=self.decoder_hidden,
                                  out_channels=num_targets,
                                  num_layers=self.decoder_num_layers,
                                  dropout=self.dropout,
                                  Normalization=self.NormLayer,
                                  InputNorm=False)
        else:
            self.V2EConvs.append(HalfNLHconv(in_dim=num_features,
                                             hid_dim=args.MLP_hidden,
                                             out_dim=args.MLP_hidden,
                                             num_layers=args.MLP_num_layers,
                                             dropout=self.dropout,
                                             Normalization=self.NormLayer,
                                             InputNorm=self.InputNorm,
                                             heads=args.heads,
                                             attention=args.PMA))
            self.bnV2Es.append(nn.BatchNorm1d(args.MLP_hidden))
            self.E2VConvs.append(HalfNLHconv(in_dim=args.MLP_hidden,
                                             hid_dim=args.MLP_hidden,
                                             out_dim=args.MLP_hidden,
                                             num_layers=args.MLP_num_layers,
                                             dropout=self.dropout,
                                             Normalization=self.NormLayer,
                                             InputNorm=self.InputNorm,
                                             heads=args.heads,
                                             attention=args.PMA))
            self.bnE2Vs.append(nn.BatchNorm1d(args.MLP_hidden))
            for _ in range(self.All_num_layers-1):
                self.V2EConvs.append(HalfNLHconv(in_dim=args.MLP_hidden,
                                                 hid_dim=args.MLP_hidden,
                                                 out_dim=args.MLP_hidden,
                                                 num_layers=args.MLP_num_layers,
                                                 dropout=self.dropout,
                                                 Normalization=self.NormLayer,
                                                 InputNorm=self.InputNorm,
                                                 heads=args.heads,
                                                 attention=args.PMA))
                self.bnV2Es.append(nn.BatchNorm1d(args.MLP_hidden))
                self.E2VConvs.append(HalfNLHconv(in_dim=args.MLP_hidden,
                                                 hid_dim=args.MLP_hidden,
                                                 out_dim=args.MLP_hidden,
                                                 num_layers=args.MLP_num_layers,
                                                 dropout=self.dropout,
                                                 Normalization=self.NormLayer,
                                                 InputNorm=self.InputNorm,
                                                 heads=args.heads,
                                                 attention=args.PMA))
                self.bnE2Vs.append(nn.BatchNorm1d(args.MLP_hidden))
            if self.GPR:
                self.MLP = MLP(in_channels=num_features,
                               hidden_channels=args.MLP_hidden,
                               out_channels=args.MLP_hidden,
                               num_layers=args.MLP_num_layers,
                               dropout=self.dropout,
                               Normalization=self.NormLayer,
                               InputNorm=False)
                self.GPRweights = nn.Linear(self.All_num_layers+1, 1, bias=False)
                self.decoder = MLP(in_channels=args.MLP_hidden,
                                      hidden_channels=self.decoder_hidden,
                                      out_channels=num_targets,
                                      num_layers=self.decoder_num_layers,
                                      dropout=self.dropout,
                                      Normalization=self.NormLayer,
                                      InputNorm=False)
            else:
                self.decoder = MLP(in_channels=args.MLP_hidden,
                                      hidden_channels=self.decoder_hidden,
                                      out_channels=num_targets,
                                      num_layers=self.decoder_num_layers,
                                      dropout=self.dropout,
                                      Normalization=self.NormLayer,
                                      InputNorm=False)

    def reset_parameters(self):
        for layer in self.V2EConvs:
            layer.reset_parameters()
        for layer in self.E2VConvs:
            layer.reset_parameters()
        for layer in self.bnV2Es:
            layer.reset_parameters()
        for layer in self.bnE2Vs:
            layer.reset_parameters()
        self.decoder.reset_parameters()
        if self.GPR:
            self.MLP.reset_parameters()
            self.GPRweights.reset_parameters()
        if self.LearnMask:
            nn.init.ones_(self.Importance)

    def forward(self, data):

        x, edge_index, norm = data.x, data.hyperedge_index, data.norm
        if self.LearnMask:
            norm = self.Importance*norm
        cidx = edge_index[1].min()
        edge_index[1] -= cidx  # make sure we do not waste memory
        reversed_edge_index = torch.stack(
            [edge_index[1], edge_index[0]], dim=0)
        if self.GPR:
            xs = []
            xs.append(F.relu(self.MLP(x)))
            for i, _ in enumerate(self.V2EConvs):
                e = F.relu(self.V2EConvs[i](x, edge_index, norm, self.aggr)) 
                x = F.dropout(e, p=self.dropout, training=self.training)
                x = self.E2VConvs[i](x, reversed_edge_index, norm, self.aggr)
                x = F.relu(x)
                xs.append(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            x = torch.stack(xs, dim=-1)
            x = self.GPRweights(x).squeeze()
            x = self.decoder(x)
        else:
            x = F.dropout(x, p=0.2, training=self.training) # Input dropout
            for i, _ in enumerate(self.V2EConvs):
                e = F.relu(self.V2EConvs[i](x, edge_index, norm, self.aggr))
                x = F.dropout(e, p=self.dropout, training=self.training)
                x = F.relu(self.E2VConvs[i](
                    x, reversed_edge_index, norm, self.aggr))
                x = F.dropout(x, p=self.dropout, training=self.training)
            x = self.decoder(x)

        return x,e




import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import defaultdict
from dhgbench.lib_models.HNN import MLP
from torch_geometric.nn import global_max_pool,global_mean_pool
from torch import Tensor


def safe_scatter_max(src: torch.Tensor, index: torch.Tensor, dim: int = 0, dim_size: int = None):

    if dim < 0:
        dim = src.dim() + dim

    if dim_size is None:
        dim_size = index.max().item() + 1

    for _ in range(src.dim() - index.dim()):
        index = index.unsqueeze(-1)
    index = index.expand_as(src)

    out_shape = list(src.shape)
    out_shape[dim] = dim_size

    out = torch.full(out_shape, float('-inf'), dtype=src.dtype, device=src.device)

    out = out.scatter_reduce(dim=dim, index=index, src=src, reduce='amax', include_self=True)

    return out

# -------------------------------- Hypergraph Classification Aggregator ------------------------------

class HyperGPredictor(nn.Module):
    def __init__(self, encoder, num_targets, args):
        super(HyperGPredictor, self).__init__()
        
        self.args=args
        self.encoder = encoder
        
        self.pooling = args.pooling 
        self.g_embed_hidden=args.g_embed_hidden
        self.g_embed_layer=args.g_embed_layer
        self.g_embed_dropout=args.g_embed_dropout
        self.g_embed_norm = args.g_embed_norm
        
        self.classifier = MLP(in_channels=args.embedding_hidden,
                hidden_channels=self.g_embed_hidden,
                out_channels=num_targets,
                num_layers=self.g_embed_layer,
                dropout=self.g_embed_dropout,
                Normalization=self.g_embed_norm,
                InputNorm=False)
    
    def reset_parameters(self):
        self.encoder.reset_parameters()
        self.classifier.reset_parameters()
    
    def forward(self,data):
        
        n_emb,_ = self.encoder(data)

        if self.pooling == 'mean':
            g_emb = global_mean_pool(n_emb,data.batch)
        elif self.pooling == 'max':
            #g_emb = global_max_pool(n_emb,data.batch)
            dim = -1 if isinstance(n_emb, Tensor) and n_emb.dim() == 1 else -2
            g_emb = safe_scatter_max(n_emb, data.batch, dim=dim)
        
        g_emb = self.classifier(g_emb)
        
        return g_emb

# -------------------------------- Hyperedge Prediction Aggregator ------------------------------

class EdgePredictor(nn.Module):
    def __init__(self, encoder, aggregator,args):
        super(EdgePredictor, self).__init__()
        self.encoder = encoder
        self.aggregator = aggregator
        self.args=args
        self.edge_aggr = args.edge_aggr 
    
    def reset_parameters(self):
        self.encoder.reset_parameters()
        self.aggregator.reset_parameters()
        
    def encoding(self,data):
        n_embed = self.encoder(data)
        return n_embed

    def aggregate(self, nfeat, hedges, mode='Train'):
        if self.edge_aggr == 'group':
            size_groups = defaultdict(list)
            for he in hedges:
                size_groups[len(he)].append(he)

            preds = []
            for size, group in size_groups.items():
                he_feats = torch.stack([nfeat[he] for he in group])  # (num_hedges, size, feature_dim)
                pred = self.aggregator(he_feats,self.edge_aggr)
                preds.append(pred)
            
            if mode == 'Train':
                return torch.cat([seq.squeeze(-1) for seq in preds],dim=0)
            elif mode == 'Eval':
                final_preds=torch.cat([seq for seq in preds],dim=0)
                return [pred.detach() for pred in final_preds]
            else:
                raise NotImplementedError
        else:
            preds = []
            if mode == 'Train':
                for he in hedges:
                    feat = nfeat[he]
                    pred = self.aggregator(feat,self.edge_aggr)
                    preds.append(pred)
                return torch.stack(preds).squeeze()

            elif mode == 'Eval':
                for he in hedges:
                    feat = nfeat[he]
                    pred = self.aggregator(feat,self.edge_aggr)
                    preds.append(pred.detach())
                return preds
            else:
                raise NotImplementedError

class MeanAggregator(nn.Module):
    def __init__(self, args):
        super(MeanAggregator, self).__init__()

        self.e_embed_hidden=args.e_embed_hidden
        self.e_embed_layer=args.e_embed_layer
        self.e_embed_dropout=args.e_embed_dropout
        self.e_embed_norm = args.e_embed_norm
        
        self.classifier = MLP(in_channels=args.embedding_hidden,
                hidden_channels=self.e_embed_hidden,
                out_channels=1,
                num_layers=self.e_embed_layer,
                dropout=self.e_embed_dropout,
                Normalization=self.e_embed_norm,
                InputNorm=False)

    def reset_parameters(self):

        self.classifier.reset_parameters()
    
    def aggregate(self, embeddings, method='group'):
        dim=1 if method == 'group'else 0
        embedding = embeddings.mean(dim=dim).squeeze()
        return embedding
    
    def classify(self, embedding) :
#         pdb.set_trace()
        embedding = torch.linalg.norm(embedding.unsqueeze(0),dim=0)      
        return self.classifier(embedding)
    
    def forward(self, embeddings, method='group'):
        embedding = self.aggregate(embeddings,method)
        pred = self.classify(embedding)
        return pred

class MaxminAggregator(nn.Module):
    def __init__(self, args):
        super(MaxminAggregator, self).__init__()

        self.e_embed_hidden=args.e_embed_hidden
        self.e_embed_layer=args.e_embed_layer
        self.e_embed_dropout=args.e_embed_dropout
        self.e_embed_norm = args.e_embed_norm
        
        self.classifier = MLP(in_channels=args.embedding_hidden,
                hidden_channels=self.e_embed_hidden,
                out_channels=1,
                num_layers=self.e_embed_layer,
                dropout=self.e_embed_dropout,
                Normalization=self.e_embed_norm,
                InputNorm=False)

    def reset_parameters(self):

        self.classifier.reset_parameters()
    
    def aggregate(self, embeddings, method='group'):
        dim=1 if method == 'group'else 0
        max_val, _ = torch.max(embeddings, dim=dim) 
        min_val, _ = torch.min(embeddings, dim=dim)
        return max_val - min_val
    
    def classify(self, embedding):
        return F.sigmoid(self.classifier(embedding))
    
    def forward(self, embeddings,method):
        embedding = self.aggregate(embeddings,method)
        pred = self.classify(embedding)
        return pred

class MaxAggregator(nn.Module):
    def __init__(self, args):
        super(MaxAggregator, self).__init__()

        self.e_embed_hidden=args.e_embed_hidden
        self.e_embed_layer=args.e_embed_layer
        self.e_embed_dropout=args.e_embed_dropout
        self.e_embed_norm = args.e_embed_norm
        
        self.classifier = MLP(in_channels=args.embedding_hidden,
                hidden_channels=self.e_embed_hidden,
                out_channels=1,
                num_layers=self.e_embed_layer,
                dropout=self.e_embed_dropout,
                Normalization=self.e_embed_norm,
                InputNorm=False)

    def reset_parameters(self):

        self.classifier.reset_parameters()
    
    def aggregate(self, embeddings, method='group'):
        dim=1 if method == 'group'else 0
        max_val, _ = torch.max(embeddings, dim=dim) 
        return max_val
    
    def classify(self, embedding):
        return F.sigmoid(self.classifier(embedding))
    
    def forward(self, embeddings,method):
        embedding = self.aggregate(embeddings,method)
        pred = self.classify(embedding)
        return pred
